import aiohttp_jinja2

from riegocloud.web.security import raise_permission

class Home():
    def __init__(self, app):
        app.router.add_get('/', self.index, name='home')


    @aiohttp_jinja2.template('home/index.html')
    async def index(self, request):
        await raise_permission(request, "superuser")
        text = "Test text"
        return {'text': text}




import asyncio
import asyncssh
import sys

from sqlite3 import IntegrityError
from riegocloud.db import get_db

from logging import getLogger
_log = getLogger(__name__)

_instance = None


def get_ssh():
    global _instance
    return _instance


def setup_ssh(app, options=None, db=None):
    global _instance
    if _instance is not None:
        del _instance
    _instance = Ssh(app, options=options, db=db)
    return _instance


class Ssh:
    def __init__(self, app, options=None, db=None):
        global _instance
        if _instance is None:
            _instance = self
        self._server = None
        app.cleanup_ctx.append(self._sshd_engine)

        # asyncssh.set_log_level(logging.DEBUG)
        # asyncssh.set_debug_level(3)

    async def _sshd_engine(self, app):
        task = asyncio.create_task(self._start_server(app))
        yield
        await self._shutdown_server(app, task)

    async def _start_server(self, app):
        self._server = await asyncssh.listen(app['options'].ssh_server_bind_address,
                                             app['options'].ssh_server_bind_port,
                                             server_host_keys=app['options'].ssh_host_key,
                                             process_factory=self._handle_client,
                                             server_factory=MySSHServer)

    async def _shutdown_server(self, app, task):
        MySSHServer._shutdown_all()
        if self._server is not None:
            self._server.close()
            await self._server.wait_closed()

    def _handle_client(self, process):
        process.stdout.write('Welcome to my SSH server, %s!\n' %
                             process.get_extra_info('username'))
        process.exit(0)


class MySSHServer(asyncssh.SSHServer):
    _conn_list = []

    @classmethod
    def _shutdown_all(cls):
        for conn in cls._conn_list:
            conn.close()

    def __init__(self):
        self._cloud_identifier = None
        self._listen_port = None
        self._conn = None

    def server_requested(self, listen_host, listen_port):
        if listen_host != 'localhost' or listen_port != self._listen_port:
            _log.error(
                f'{self._cloud_identifier}: unallowed TCP forarding requested')
            return False
        _log.debug(f'Port forwarding established: {listen_host}:{listen_port}')
        return True

    def connection_made(self, conn):
        self._conn = conn
        _log.debug('SSH connection received from {}'.format(
            conn.get_extra_info('peername')[0]))
        MySSHServer._conn_list.append(conn)

    def connection_lost(self, exc):
        if exc:
            _log.debug(f'SSH connection error: {exc}')
        else:
            _log.debug('SSH connection closed.')
        MySSHServer._conn_list.remove(self._conn)
        self._conn = None

    def begin_auth(self, username):
        cursor = get_db().conn.cursor()
        cursor.execute("""SELECT *
                          FROM clients
                          WHERE cloud_identifier = ?""", (username,))
        row = cursor.fetchone()
        if row is None or row['is_disabled']:
            return True

        self._cloud_identifier = username
        self._listen_port = row['ssh_server_listen_port']
        print(row['public_user_key'])
        key = asyncssh.import_authorized_keys(
            row['public_user_key'])
        self._conn.set_authorized_keys(key)
        return True

    def public_key_auth_supported(self):
        return True

    def validate_public_key(self, username, key):
        return True



import sqlite3
import asyncio
from pathlib import Path
from yoyo import read_migrations
from yoyo import get_backend

from logging import getLogger
_log = getLogger(__name__)

_instance = None


def get_db():
    global _instance
    return _instance


def setup_db(options=None):
    global _instance
    if _instance is not None:
        del _instance
    _instance = Db(options=options)
    return _instance


class Db:
    def __init__(self, options=None):
        global _instance
        if _instance is None:
            _instance = self

        self.conn = None
        Path(options.db_filename).parent.mkdir(
            parents=True, exist_ok=True)

        self._do_migrations(options)

        try:
            self.conn = sqlite3.connect(options.db_filename,
                                        detect_types=sqlite3.PARSE_DECLTYPES)
        except Exception as e:
            _log.error(f'Unable to connect to database: {e}')
            if self.conn is not None:
                self.conn.close()
            exit(1)
        self.conn.execute("PRAGMA foreign_keys = ON")
        self.conn.row_factory = sqlite3.Row

    def _do_migrations(self, options):
        try:
            backend = get_backend('sqlite:///' + options.db_filename)
        except Exception as e:
            _log.error(f'Unable to open database: {e}')
            exit(1)

        migrations = read_migrations(options.db_migrations_dir)
        with backend.lock():
            backend.apply_migrations(backend.to_apply(migrations))

    def __del__(self):
        try:
            if self.conn is not None:
                self.conn.close()
        except Exception as e:
            _log.error(f'database.py: Unable to close Databse: {e}')



from jinja2 import FileSystemLoader, Environment, TemplateNotFound
import asyncio
from pathlib import Path


from logging import getLogger
_log = getLogger(__name__)

async def create_apache_conf(options=None):

    env = Environment(
        loader=FileSystemLoader(Path(options.apache_tpl_file).parent),
        autoescape=False
    )
    try: 
        template=env.get_template(Path(options.apache_tpl_file).name)
    except TemplateNotFound as e:
        _log.error(f'No template found: {e}')
        return False

    Path(options.apache_conf_file).parent.mkdir(
            parents=True, exist_ok=True)

    with open(options.apache_conf_file, "w") as f:
        f.write(template.render(clients=clients))

    try:
        proc = await asyncio.create_subprocess_exec(
            '/usr/bin/sudo', 
            '/usr/sbin/apachectl', 'graceful',
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)
        stdout, stderr = await proc.communicate()
        print(f'stdout: {stdout}, stderr: {stderr}')
    except FileNotFoundError as e:
        print(f'Apache not reloaded: {e}')

    
    

        
class Options():

    def __init__(self):
        self.apache_tpl_file = "riegocloud/apache/apache.conf.tpl"
        self.apache_conf_file = "riegocloud/apache/apache.conf"

clients = [
    {'ssh_server_listen_port' : '33337', 'cloud_identifier' : 'ident1'},
    {'ssh_server_listen_port' : '33338', 'cloud_identifier' : 'ident2'},
    {'ssh_server_listen_port' : '33338', 'cloud_identifier' : 'ident3'},
]

async def main():
    app = {}
    options = Options()
    app['options'] = options

    await create_apache_conf(options=options)


if __name__ == "__main__":
    asyncio.run(main())


import aiohttp_jinja2
from aiohttp import web

from riegocloud.web.security import raise_permission

import asyncio
import sys
import json

from pkg_resources import packaging

router = web.RouteTableDef()


def setup_routes_system(app):
    app.add_routes(router)


@router.get("/system", name='system')
@aiohttp_jinja2.template('system/index.html')
async def system_index(request):
    await raise_permission(request, permission=None)
    text = ''
    return {"text": text}


@router.get("/system/check_update", name='system_check_update')
@aiohttp_jinja2.template('system/index.html')
async def system_check_update(request):
    await raise_permission(request, permission=None)
    update = await _check_update("No update")
    return {'text':  update}


@router.get("/system/do_update", name='system_do_update')
@aiohttp_jinja2.template('system/index.html')
async def system_do_update(request):
    await raise_permission(request, permission=None)
    await _do_update()
    return {'text': "Restart erforderlich"}


@router.get("/system/restart", name='system_restart')
@aiohttp_jinja2.template('system/index.html')
async def system_restart(request):
    await raise_permission(request, permission=None)
    # TODO shedule exit for a few seconds and return a redirect
    asyncio.get_event_loop().call_later(1, exit, 0)
    raise web.HTTPSeeOther(request.app.router['system'].url_for())


async def _check_installed():
    version = "0.0.0"
    proc = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'pip', 'list', "--format=json",
        "--disable-pip-version-check",
        "--no-color",
        "--no-python-version-warning",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE)

    stdout, stderr = await proc.communicate()
    data = json.loads(stdout)
    for item in data:
        if item['name'] == 'riegocloud':
            version = item['version']
            break
    return version


async def _check_update(latest_version=None):
    proc = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'pip', 'list', "-o", "--format=json",
        "--disable-pip-version-check",
        "--no-color",
        "--no-python-version-warning",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE)

    stdout, stderr = await proc.communicate()
    data = json.loads(stdout)
    for item in data:
        if item['name'] == 'riegocloud':
            latest_version = item['latest_version']
            break
    return latest_version


async def _do_update():
    proc = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'pip', 'install', "riegocloud", "--upgrade",
        "--disable-pip-version-check",
        "--no-color",
        "--no-python-version-warning",
        "-q", "-q", "-q")
    await proc.wait()
    exit (0)
    #return proc.returncode



from aiohttp import web
import asyncio
import sys
import json
import asyncssh
from jinja2 import FileSystemLoader, Environment, TemplateNotFound
from pathlib import Path

from sqlite3 import IntegrityError
from riegocloud.db import get_db

from logging import getLogger
_log = getLogger(__name__)

router = web.RouteTableDef()


def setup_routes_api(app):
    app.add_routes(router)


@router.post("/api_20210221/", name='api')
async def api_post(request):
    options = request.app['options']
    data = await request.json()
    cloud_identifier = data.get('cloud_identifier', '')
    public_user_key = data.get('public_user_key', '')
    if not len(cloud_identifier) or not len(public_user_key):
        await asyncio.sleep(5)
        raise web.HTTPBadRequest

    cursor = get_db().conn.cursor()
    cursor.execute("""SELECT MAX(ssh_server_listen_port) AS max_port
                    FROM clients""")
    max_port = cursor.fetchone()
    max_port = max_port['max_port']
    if max_port is None or max_port == 0:
        ssh_server_listen_port = 33333
    else:
        ssh_server_listen_port = max_port + 1
    try:
        with get_db().conn:
            get_db().conn.execute(
                """INSERT INTO clients
                    ('cloud_identifier', 'public_user_key', ssh_server_listen_port,
                    ssh_server_hostname, ssh_server_port)
                    VALUES (?,?,?,?,?)""",
                (cloud_identifier, public_user_key, ssh_server_listen_port,
                 options.ssh_server_hostname, options.ssh_server_port))
    except IntegrityError:
        try:
            with get_db().conn:
                get_db().conn.execute(
                    '''UPDATE clients SET
                       public_user_key = ?,
                       ssh_server_listen_port = ?,
                       ssh_server_hostname = ?,
                       ssh_server_port = ?
                       WHERE cloud_identifier = ? ''',
                    (public_user_key, ssh_server_listen_port,
                     options.ssh_server_hostname, options.ssh_server_port,
                     cloud_identifier))
        except IntegrityError as e:
            _log.error(f'Cannot update Client: {e}')
            await asyncio.sleep(5)
            raise web.HTTPBadRequest
        else:
            _log.debug(f'Updated Client: {cloud_identifier}')
    else:
        _log.debug(f'Created Client: {cloud_identifier}')

    data['ssh_server_hostname'] = options.ssh_server_hostname
    data['ssh_server_port'] = options.ssh_server_port
    data['ssh_server_listen_port'] = ssh_server_listen_port

    await create_apache_conf(options=options)
    return web.json_response(data)


async def create_apache_conf(options=None):
    cursor = get_db().conn.cursor()
    cursor.execute("SELECT * FROM clients")
    clients = cursor.fetchall()

    env = Environment(
        loader=FileSystemLoader(Path(options.apache_tpl_file).parent),
        autoescape=False
    )
    try: 
        template=env.get_template(Path(options.apache_tpl_file).name)
    except TemplateNotFound as e:
        _log.error(f'No template found: {e}')
        return False

    with open(options.apache_conf_file, "w") as f:
        f.write(template.render(clients=clients))

    try:
        proc = await asyncio.create_subprocess_exec(
            '/usr/bin/sudo', 
            '/usr/sbin/apachectl', 'graceful',
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)
        stdout, stderr = await proc.communicate()
        print(f'stdout: {stdout}, stderr: {stderr}')
    except FileNotFoundError as e:
        print(f'Apache not reloaded: {e}')
        return False
    return True


import asyncio
import configargparse
import pkg_resources
import os
import sys
from pathlib import Path

import logging
from logging.handlers import RotatingFileHandler

from aiohttp import web
import aiohttp_jinja2
import jinja2
import aiohttp_debugtoolbar
from aiohttp_remotes import setup as setup_remotes, XForwardedRelaxed
from aiohttp_session import setup as session_setup, get_session
from aiohttp_session.memcached_storage import MemcachedStorage
import aiomcache


from riegocloud.web.views.home import Home
from riegocloud.web.views.system import setup_routes_system
from riegocloud.web.views.api import setup_routes_api
from riegocloud.web.security import current_user_ctx_processor, setup_routes_security

from riegocloud.ssh import setup_ssh
from riegocloud.db import setup_db

from riegocloud import __version__

app_name = 'riegocloud'


async def on_startup(app):
    logging.getLogger(__name__).debug("on_startup")


async def on_shutdown(app):
    logging.getLogger(__name__).debug("on_shutdown")


async def on_cleanup(app):
    logging.getLogger(__name__).debug("on_cleanup")


def main():
    options = _get_options()

    _setup_logging(options=options)

    if sys.version_info >= (3, 8) and options.WindowsSelectorEventLoopPolicy:
        asyncio.DefaultEventLoopPolicy = asyncio.WindowsSelectorEventLoopPolicy  # noqa: E501

    if os.name == "posix":
        import uvloop  # pylint: disable=import-error
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    web.run_app(run_app(options=options),
                host=options.http_server_bind_address,
                port=options.http_server_bind_port)


async def run_app(options=None):
    loop = asyncio.get_event_loop()

    if options.enable_asyncio_debug_log:
        loop.set_debug(True)

    app = web.Application()

    app.on_startup.append(on_startup)
    app.on_shutdown.append(on_shutdown)
    app.on_cleanup.append(on_cleanup)

    app['version'] = __version__
    app['options'] = options

    mcache = aiomcache.Client(options.memcached_host, options.memcached_port)
    session_setup(app, MemcachedStorage(mcache))

    async def mcache_shutdown(app):
        await mcache.close()
    app.on_shutdown.append(mcache_shutdown)

    loader = jinja2.FileSystemLoader(options.http_server_template_dir)
    aiohttp_jinja2.setup(app,
                         loader=loader,
                         # enable_async=True,
                         context_processors=[current_user_ctx_processor],
                         )

    await setup_remotes(app, XForwardedRelaxed())
    if options.enable_aiohttp_debug_toolbar:
        aiohttp_debugtoolbar.setup(
            app, check_host=False, intercept_redirects=False)

    app.router.add_static('/static', options.http_server_static_dir,
                          name='static', show_index=True)

    db = setup_db(options=options)
    setup_ssh(app, options=options, db=db)
    setup_routes_security(app)
    setup_routes_system(app)
    setup_routes_api(app)

    Home(app)

    return app


def _setup_logging(options=None):
    formatter = logging.Formatter(
        "%(asctime)s;%(levelname)s;%(name)s;%(message)s ")
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    if options.verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO
    Path(options.log_file).parent.mkdir(parents=True, exist_ok=True)
    file_handler = RotatingFileHandler(options.log_file, mode='a',
                                       maxBytes=options.log_max_bytes,
                                       backupCount=options.log_backup_count,
                                       encoding=None, delay=0)
    file_handler.setFormatter(formatter)
    logging.basicConfig(level=level, handlers=[stream_handler, file_handler])

    if options.enable_aiohttp_access_log:
        logging.getLogger("aiohttp.access").setLevel(logging.DEBUG)
    else:
        logging.getLogger("aiohttp.access").setLevel(logging.ERROR)

    if options.enable_ssh_debug_log:
        logging.getLogger("asyncssh").setLevel(logging.DEBUG)
    else:
        logging.getLogger("asyncssh").setLevel(logging.ERROR)


def _get_options():
    p = configargparse.ArgParser(
        default_config_files=[f'/etc/{app_name}/conf.d/*.conf',
                              f'~/.{app_name}.conf',
                              f'{app_name}.conf'],
        args_for_writing_out_config_file=['-w',
                                          '--write-out-config-file'])
    p.add('-c', '--config', is_config_file=True, env_var=app_name.capitalize(),
          required=False, help='config file path')
# Database
    p.add('-d', '--db_filename', help='Path and name for DB file',
          default=f'db/{app_name}.db')
    p.add('--db_migrations_dir',
          help='path to database migrations directory',
          default=pkg_resources.resource_filename(app_name, 'migrations'))
# Logging
    p.add('-l', '--log_file', help='Full path to logfile',
          default=f'log/{app_name}.log')
    p.add('--log_max_bytes', help='Maximum logfile size in bytes',
          default=1024*300, type=int)
    p.add('--log_backup_count', help='How many files to rotate',
          default=3, type=int)
# Secrets & Scurity
    p.add('--max_age_remember_me', type=int, default=7776000)
    p.add('--cookie_name_remember_me', default="remember_me")
    p.add('--reset_admin', help='Reset admin-pw to given value an exit')
# Memcache
    p.add('--memcached_host', help='IP adress of memcached host',
          default='127.0.0.1')
    p.add('--memcached_port', help='Port of memcached service',
          default=11211, type=int)
# HTTP-Server / API
    p.add('--http_server_bind_address',
          help='http-server bind address', default='127.0.0.1')
    p.add('--http_server_bind_port', help='http-server bind port',
          default=8181, type=int)
    p.add('--ssh_server_hostname', help='Send this hostname to client',
          default="cloud.finca-panorama.es")
    p.add('--ssh_server_port', help='Send this port to client',
          default=8022, type=int)
# SSH-Server
    p.add('--ssh_server_bind_port', help='ssh-server bind port',
          default=8022, type=int)
    p.add('--ssh_server_bind_address', help='ssh-server bind address',
          default='0.0.0.0')
    p.add('--ssh_host_key',
          help='ssh Host key', default='ssh/ssh_host_key')
# Apache patching
    p.add('--apache_tpl_file', help='path to apache config template(s)',
          default=pkg_resources.resource_filename(
              app_name, 'apache/apache.conf.tpl'))
    p.add('--apache_conf_file', help='path to apache config file',
          default='apache/apache.conf')

# Directories
    p.add('--base_dir', help='Change only if you know what you are doing',
          default=Path(__file__).parent)
    p.add('--http_server_static_dir',
          help='Serve static html files from this directory',
          default=pkg_resources.resource_filename(f'{app_name}.web', 'static'))
    p.add('--http_server_template_dir',
          help='Serve template files from this directory',
          default=pkg_resources.resource_filename(f'{app_name}.web', 'templates'))
# Debug
    p.add('--enable_aiohttp_debug_toolbar', action='store_true')
    p.add('--enable_aiohttp_access_log', action='store_true')
    p.add('--enable_asyncio_debug_log', action='store_true')
    p.add('--enable_ssh_debug_log', action='store_true')
    p.add('--WindowsSelectorEventLoopPolicy', action='store_true')

# Version, Help, Verbosity
    p.add('-v', '--verbose', help='verbose', action='store_true')
    p.add('--version', help='Print version and exit', action='store_true')
    p.add('--defaults', help='Print options with default values and exit',
          action='store_true')

    options = p.parse_args()
    if options.verbose:
        print(p.format_values())

    try:
        with open(f'{app_name}.conf', 'xt') as f:
            for item in vars(options):
                f.write(f'# {item}={getattr(options, item)}\n')
    except IOError:
        pass

    if options.defaults:
        for item in vars(options):
            print(f'# {item}={getattr(options, item)}')
        exit(0)

    if options.version:
        print('Version: ', __version__)
        exit(0)

    if options.reset_admin:
        _reset_admin(options)
        exit(0)

    return options


def _reset_admin(options):
    from sqlite3 import IntegrityError
    from riegocloud.db import setup_db
    import bcrypt

    db = setup_db(options=options)
    password = options.reset_admin

    if len(password) > 0:
        password = password.encode('utf-8')
        password = bcrypt.hashpw(password, bcrypt.gensalt(12))
        try:
            with db.conn:
                db.conn.execute(
                    '''UPDATE users
                        SET password = ?
                        WHERE id = ? ''', (password, 1))
        except IntegrityError as e:
            print(f'Unable to reset Admin PW: {e}')
        else:
            print(f'Succesfully reset Admin PW: {password}')
    db.conn.close()



from aiohttp_session import get_session
import aiohttp_jinja2
from sqlite3 import IntegrityError, Row
from riegocloud.db import get_db
from aiohttp import web
import bcrypt
import secrets
import json
import asyncio

from logging import getLogger
_log = getLogger(__name__)


async def current_user_ctx_processor(request):
    user = await get_user(request)
    return {'user': user}

router = web.RouteTableDef()


def setup_routes_security(app):
    app.add_routes(router)


async def get_user(request) -> Row:
    session = await get_session(request)
    db = get_db()
    user_id = session.get('user_id')
    if user_id is not None:
        cursor = db.conn.cursor()
        cursor.execute("""SELECT *, 'login' AS 'provider'
                          FROM users
                          WHERE id = ?""", (user_id,))
        user = cursor.fetchone()
        if user is None or user['is_disabled']:
            session.pop('user_id', None)
            return None
        return user
    remember_me = request.cookies.get('remember_me')
    if remember_me is not None:
        cursor = db.conn.cursor()
        cursor.execute("""SELECT *, 'cookie' AS 'provider'
                           FROM users
                           WHERE remember_me = ?""", (remember_me,))
        user = cursor.fetchone()
        if user is not None and user['is_disabled']:
            try:
                with db.conn:
                    db.conn.execute("""UPDATE users
                                       SET remember_me = ''
                                       WHERE id = ?""", (user['id'],))
            except IntegrityError:
                pass
            return None
        return user
    return None


async def check_permission(request, permission=None) -> Row:
    user = await get_user(request)
    db = get_db()
    if user is None:
        return None
    if user['is_disabled']:
        return None
    if user['is_superuser']:
        return user
    if permission is None:
        return user

    cursor = db.conn.cursor()
    cursor.execute(
        'SELECT * FROM users_permissions WHERE user_id = ?', (user['id'],))
    for row in cursor:
        if row['name'] == permission:
            return user
    return None


async def raise_permission(request: web.BaseRequest, permission: str = None):
    """Generate redirection to login form if permission is not
    sufficent. Append query string with information for redirecting
    after login to the original url.

    :param request: [description]
    :type request: web.Baserequest
    :param permission: If no permission is given, check auth only
    :type permission: str, optional
    :raises web.HTTPSeeOther: [description]
    """
    if await check_permission(request, permission=permission) is None:
        raise web.HTTPSeeOther(
            request.app.router['login'].url_for(
            ).with_query(
                {"redirect": str(request.rel_url)}
            )
        )


@router.get("/login", name='login')
@aiohttp_jinja2.template("security/login.html")
async def _login(request: web.Request):
    redirect = request.rel_url.query.get("redirect", "")
    csrf_token = secrets.token_urlsafe()
#    session = await new_session(request)
    session = await get_session(request)
    session['csrf_token'] = csrf_token
    return {'csrf_token': csrf_token, 'redirect': redirect}


@ router.post("/login")
async def _login_apply(request: web.Request):
    form = await request.post()
    session = await get_session(request)
    if session.get('csrf_token') != form['csrf_token']:
        # Normally not possible
        await asyncio.sleep(2)
        raise web.HTTPUnauthorized()

    if form.get('identity') is None:
        await asyncio.sleep(2)
        raise web.HTTPSeeOther(request.app.router['login'].url_for())

    cursor = get_db().conn.cursor()
    cursor.execute("""SELECT *, 'login' AS 'provider'
                    FROM users
                    WHERE identity = ?""", (form['identity'],))
    user = cursor.fetchone()

    if (
        user is None or
        user['is_disabled'] or
        user['password'] is None or
        not len(user['password'])
    ):
        await asyncio.sleep(2)
        raise web.HTTPSeeOther(request.app.router['login'].url_for())
    if not bcrypt.checkpw(form['password'].encode('utf-8'), user['password']):
        await asyncio.sleep(2)
        raise web.HTTPSeeOther(request.app.router['login'].url_for())

    session['user_id'] = user['id']

    location = form.get('redirect')
    if location is None or location == '':
        location = request.app.router['home'].url_for()
    response = web.HTTPSeeOther(location=location)
    if form.get('remember_me') is not None:
        remember_me = secrets.token_urlsafe()
        try:
            with get_db().conn:
                get_db().conn.execute(
                    '''UPDATE users
                       SET remember_me = ?
                       WHERE id = ? ''',
                    (remember_me, user['id']))
        except IntegrityError:
            pass
        response.set_cookie("remember_me", remember_me,
                            max_age=request.app['options'].max_age_remember_me,
                            httponly=True,
                            samesite='strict')
    return response


@ router.get("/logout", name='logout')
async def _logout(request: web.Request):
    user = await get_user(request)
    if user is not None:
        try:
            with get_db().conn:
                get_db().conn.execute("""UPDATE users
                                        SET remember_me = ''
                                        WHERE id = ?""", (user['id'],))
        except IntegrityError:
            pass
    session = await get_session(request)
    if session is not None:
        session.pop('user_id', None)
    response = web.HTTPSeeOther(request.app.router['login'].url_for())
#    response.set_cookie('remember_me', None,
#                        expires='Thu, 01 Jan 1970 00:00:00 GMT')
    response.del_cookie('remember_me')
    return response


@ router.get("/passwd", name='passwd')
@ aiohttp_jinja2.template("security/passwd.html")
async def _passwd(request: web.Request):
    return {}


@ router.post("/passwd")
async def _passwd_apply(request: web.Request):
    form = await request.post()
    user = await get_user(request)

# TODO check old_password and equality of pw1 an pw2
    password = form['new_password_1'].encode('utf-8')
    password = bcrypt.hashpw(password, bcrypt.gensalt(12))
    try:
        with get_db().conn:
            get_db().conn.execute(
                '''UPDATE users
                    SET password = ?
                    WHERE id = ? ''', (password, user['id']))
    except IntegrityError:
        raise web.HTTPSeeOther(request.app.router['passwd'].url_for())
    else:
        raise web.HTTPSeeOther(request.app.router['home'].url_for())
    return {}  # not reached
